#! /usr/bin/python
# File: addrbook.py
# Author: Don Stokes <don@stokesautomation.com>
# Purpose:
#  This is a simple address book utility.
#  It uses SQLite3 for storage, making it easy to manually manipulate
#  the database (be careful). Try sqlitebrowser.
#  It has a command-line console interface and a GTK GUI (in development).
#  Run the application with the -help option for usage information.
#  Each "Entity" (a person or organization) has names (Claus, Santa),
#  phone numbers, email addresses, and geographical addresses (soon).
#  Each of the phone numbers and addresses has a "role" associated with it.
#  A role can be any string, like "work", "home", "mobile-personal".
#  These items also have an integer "priority" with zero being greatest.
#  The intention is each class of data belonging to an Entity would have
#  one instance with a priority of zero.  This would be the preferred item,
#  such as preferred email address.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ToDo:
# * Add flags for dirty fields that need to be updated in the database.
# * Allow search with multiple name components
# * Add ad hoc SQL queries
# * Add physical addresses
# *  Add Google map image support
# * Add mass email feature
# * Support update
#===============================================================================
# Typical python modules
import sys
import os
import traceback
# SQLite module
import sqlite3
# Tk module (to be removed in favor of gtk)
import Tkinter
# GTK module stuff
import pygtk
import gtk
#===============================================================================
class Name:
	# A single component of a proper name - like "Santa" or "Claus".
	# Priority is used to sort them:
	#  santa.priority = 1; clause.priority = 0
	# idEntity is the foreign key to the Entity table. See Entity class.
	# Future enhancement includes adding a displayPriority for a natural appearance:
	#  "Santa" "Claus"
	# A key of -1 means the item has is new and has not been stored in the database
	# yet.  If the key is non-negative, a corresponding record exists in the database.
	def __init__(self, id=-1, value=None, priority=0, idEntity=-1):
		self.id = id
		self.value = value
		self.priority = priority
		self.idEntity = idEntity
	def updateDb(self, ab):
		if self.id < 0:
			# Insert
			cursor = ab.conn.cursor()
			value = self.value.replace("'", "''") # escape quote chars
			sql = "INSERT INTO Name(value,priority,idEntity)" \
			    + "VALUES('"+value+"',"+str(self.priority)+  ","+str(self.idEntity)+  ")"
#			print sql
			cursor.execute(sql)
			ab.conn.commit
			self.id = cursor.lastrowid
		else:
			# Update
			pass # FIXTHIS - todo
#===============================================================================
class Email:
	# This is an email address associated with an entity.
	# Objects of this class contain a the email address, role (i.e. work or home),
	# and a priority (0 is greatest).  The preferred email should have a priority
	# of 0 and others should have higher integers (making them lower priority).
	def __init__(self, id=-1, value=None, role=None, priority=1, idEntity=-1):
		self.id = id
		self.value = value
		self.role = role
		self.priority = priority
		self.idEntity = idEntity
	def toString(self):
		ret = self.value
		if self.role != None:
			ret += " (" + self.role + ")"
		return ret
	def updateDb(self, ab):
		if self.id < 0:
			# Insert
			cursor = ab.conn.cursor()
			value = self.value.replace("'", "''") # escape quote chars
			role = self.role.replace("'", "''") # escape quote chars
			sql = "INSERT INTO Email(value,role,priority,idEntity)" \
			    + "VALUES('"+value+"','" +  role + "',"   +str(self.priority)+  ","+str(self.idEntity)+  ")"
#			print sql
			cursor.execute(sql)
			ab.conn.commit
			self.id = cursor.lastrowid
		else:
			# Update
			pass # FIXTHIS
#===============================================================================
class Phone:
	# This is a phone number associated with an entity.
	# Objects of this class contain a the phone number, role (i.e. work), and
	# a priority (0 is greatest).
	def __init__(self, id=-1, value=None, role=None, priority=1, idEntity=-1):
		# Class method for constructing object with specifics
		self.id = id
		self.value = value
		self.role = role
		self.priority = priority
		self.idEntity = idEntity
	def toString(self):
		ret = self.value
		if self.role != None:
			ret += " (" + self.role + ")"
		return ret
	def updateDb(self, ab):
		if self.id < 0:
			# Insert
			cursor = ab.conn.cursor()
			value = self.value.replace("'", "''") # escape quote chars
			role = self.role.replace("'", "''") # escape quote chars
			sql = "INSERT INTO Phone(value,role,priority,idEntity)" \
			    + "VALUES('"+value+"','" +  role + "',"   +str(self.priority)+  ","+str(self.idEntity)+  ")"
#			print sql
			cursor.execute(sql)
			ab.conn.commit
			self.id = cursor.lastrowid
		else:
			# Update
			pass # FIXTHIS
#===============================================================================
class Entity:
	# An Entity is a person or an organization.
	# An Entity has 0 or more Names, ordered by priority, i.e. Claus, Santa.
	# An Entity also has phone numbers and email addresses.
	# Coming soon: Geographical addresses, i.e. 131313 Mockingbird Ln, Detroit, MI
#-------------------------------------------------------------------------------
	def __init__(self):
		# Constructor: Initialize all member data to reasonable defaults.
		self.id = -1
		self.comment = None
		self.priority = 0
		self.names = []
		self.emails = []
		self.phones = []
#-------------------------------------------------------------------------------
	def getName(self):
		ret = ""
		if self.names != None:
			cnt = len(self.names)
			i = 0
			while i < cnt:
				if i == 1:
					ret += ", "
				elif i > 1:
					ret += " "
				ret += self.names[i].value
				i += 1
		return ret
#-------------------------------------------------------------------------------
	def importVcardName(self, strN):
		parts = strN.partition(":") # Split tag and values
		values = parts[2]
		nameCnt = 0
		curName = ""
		valIndex = 0
		valLen = len(values)
		while valIndex < valLen:
			if values[valIndex] == '\\': # escape next char
				valIndex += 1
				curName += values[valIndex]
			elif values[valIndex] == ';': # separator
				if len(curName) > 0:
					self.names.append(Name(-1,curName,nameCnt,self.id))
					nameCnt += 1
					curName = ""
			else:
				curName += values[valIndex]
				
			valIndex += 1
		# Finish any unfinished name
		if len(curName) > 0:
			self.names.append(Name(-1,curName,nameCnt,self.id))
#-------------------------------------------------------------------------------
	def importVcardPhone(self, strTel):
		parts = strTel.partition(":") # Split tag and value
		tag = parts[0]
		value = parts[2]
		role = "home"
		if tag.find("work") >= 0:
			role = "work"
		phType = None
		if tag.find("cell") >= 0:
			phType = "mobile"
		if phType != None:
			role += "-" + phType
		phone = Phone(id=-1, value=value, role=role, priority=1, idEntity=-1)
		self.phones.append(phone)
#-------------------------------------------------------------------------------
	def updateDb(self, ab):
		cursor = ab.conn.cursor()
		if self.id < 0:
			# Insert
			cursor = ab.conn.cursor()
			comment = self.comment.replace("'", "''") # escape quote chars
			sql = "INSERT INTO Entity(comment,priority)" \
			    + "VALUES('"+comment+"',"+str(self.priority)+")"
#			print sql
			cursor.execute(sql)
			self.id = cursor.lastrowid
			# Insert the names
			for name in self.names:
				name.idEntity = self.id  # Populate the foreign key
				name.updateDb(ab)
			# Insert the email addresses
			for email in self.emails:
				email.idEntity = self.id  # Populate the foreign key
				email.updateDb(ab)
			# Insert the phone number
			for phone in self.phones:
				phone.idEntity = self.id  # Populate the foreign key
				phone.updateDb(ab)
		else:
			# Update
			pass		
#===============================================================================
class AddrBook:
#-------------------------------------------------------------------------------
	def __init__(self):
		self.dbFileName = "addrbook.sqlite3"
		self.conn = None
#-------------------------------------------------------------------------------
	def __del__(self):
		self.disconnect()
#-------------------------------------------------------------------------------
	def connect(self):
		fileExists = os.path.exists(self.dbFileName)
		self.conn = sqlite3.connect(self.dbFileName)
		if  not fileExists:
			self.createTables()
#-------------------------------------------------------------------------------
	def disconnect(self):
		if self.conn != None:
			self.conn.close()
			self.conn = None
#-------------------------------------------------------------------------------
	def createTables(self):
		stmnts = (
			"CREATE TABLE Entity (id INTEGER PRIMARY KEY, comment TEXT, priority NUMERIC)",
			"CREATE TABLE Name   (id INTEGER PRIMARY KEY, idEntity NUMERIC, priority NUMERIC, value TEXT)",
			"CREATE TABLE Email  (id INTEGER PRIMARY KEY, idEntity NUMERIC, priority NUMERIC, value TEXT, role TEXT);",
			"CREATE TABLE Phone  (id INTEGER PRIMARY KEY, idEntity NUMERIC, priority NUMERIC, value TEXT, role TEXT);"
		)
		cursor = self.conn.cursor()
		for stmnt in stmnts:
			cursor.execute(stmnt)
		self.conn.commit()
#-------------------------------------------------------------------------------
	def searchByName(self, name):
		result = []
		cursor = self.conn.cursor()
		strSql = "SELECT idEntity FROM Name WHERE value ='" + name +"'"
		cursor.execute(strSql)
		for row in cursor:
			result.append(self.getEntity(row[0]))
		return result
#-------------------------------------------------------------------------------
	def getNames(self, id):
		names = []
		cursor = self.conn.cursor()
		strSql = "SELECT value, priority FROM Name WHERE idEntity = " \
		       + str(id) + " ORDER BY priority"
		cursor.execute(strSql)
		for row in cursor:
			name = Name()
			name.value = row[0]
			names.append(name)
		return names
#-------------------------------------------------------------------------------
	def getAllNames(self):
		names = {}
		cursor = self.conn.cursor()
		strSql = "SELECT idEntity, value, priority FROM Name ORDER BY priority"
		cursor.execute(strSql)
		for row in cursor:
			key = row[0]
			value = row[1]
			if names.has_key(key):
				names[key].append(value)
			else:
				lst = [ value ]
				names[key] = lst
		return names
#-------------------------------------------------------------------------------
	def getEmails(self, id):
		emails = []
		cursor = self.conn.cursor()
		strSql = "SELECT id, value, role, priority, idEntity FROM Email WHERE idEntity = " \
		       + str(id) + " ORDER BY priority"
		cursor.execute(strSql)
		for row in cursor:
			email = Email(row[0], row[1], row[2], row[3], row[4])
			emails.append(email)
		return emails
#-------------------------------------------------------------------------------
	def getPhones(self, id):
		phones = []
		cursor = self.conn.cursor()
		strSql = "SELECT id, value, role, priority, idEntity FROM Phone WHERE idEntity = " \
		       + str(id) + " ORDER BY priority"
		cursor.execute(strSql)
		for row in cursor:
			phone = Phone(row[0], row[1], row[2], row[3], row[4])
			phones.append(phone)
		return phones
#-------------------------------------------------------------------------------		
	def getEntity(self, id):
		entity = Entity()
		entity.id = id
		entity.names = self.getNames(id)
		entity.emails = self.getEmails(id)
		entity.phones = self.getPhones(id)
		return entity
#-------------------------------------------------------------------------------
	def importVcard(self, strs):
		entity = Entity()
		for s in strs:
			if s.find("FN:") == 0:
				parts = s.partition(":")
				entity.comment = parts[2];
			elif s.find("N:") == 0:
				entity.importVcardName(s)
			elif s.find("TEL") == 0:
				entity.importVcardPhone(s)
			elif s.find("EMAIL") == 0:
				parts = s.partition(":")
				entity.emails.append(Email(-1,parts[2],"personal", 0,-1))
		entity.updateDb(self)		
#-------------------------------------------------------------------------------
	def importVcards(self, fileName):
		try:
			lines = []
			f = open(fileName)
			while True:
				l = f.readline()
				if len(l) == 0:
					break
				if l.find("BEGIN:VCARD") == 0: # Start a vcard
					lines = []
				elif l.find("END:VCARD") == 0: # Finish a vcard
					self.importVcard(lines)
					lines = None
				elif lines != None: # Gathering lines for current vcard
					lines.append(l.rstrip("\n"))
			f.close()
			self.conn.commit()
		except Exception, e:
			self.conn.rollback()
			print "ERROR: import failed"
			raise e # propogate exception
#===============================================================================
class GuiGtk:
#-------------------------------------------------------------------------------
	def onDestroy(self, widget, data=None):
		gtk.main_quit()
#-------------------------------------------------------------------------------
#	def delete_event(self, widget, event, data=None):
#		return False
#-------------------------------------------------------------------------------
	def __init__(self, ab=None):
		self.ab = ab

		self.lsNames = gtk.ListStore(str, str)
		if self.ab != None:
			names = ab.getAllNames().values()
			for entNames in names:
				self.lsNames.append(GuiGtk.mkTwoCols(entNames))

		self.tr  = gtk.CellRendererText()
		self.tr2 = gtk.CellRendererText()

		self.lc = gtk.TreeViewColumn("Names")
		self.lc.pack_start(self.tr, True)
		self.lc.pack_start(self.tr2, True)
		self.lc.set_attributes(self.tr, text = 0)
		self.lc.set_attributes(self.tr2, text = 1)

		self.lbNames = gtk.TreeView(self.lsNames)
		self.lbNames.append_column(self.lc)

		self.winMain = gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.winMain.add(self.lbNames)
		self.winMain.set_title("Address Book");
		self.winMain.connect("destroy", self.onDestroy)
		self.winMain.set_size_request(245, 512)
		self.winMain.show_all()

		self.run()
#-------------------------------------------------------------------------------
	def mkTwoCols(strLst):
		ret = ["",""]
		strsCnt = len(strLst)
		if strsCnt > 0:
			ret[0] = strLst[0]
			secCol = ""
			i = 1
			while i < strsCnt:
				if len(secCol) > 0:
					secCol += " "
				secCol += strLst[i]
				i += 1
			ret[1] = secCol
#			print ret[0], ret[1]
		return ret
	mkTwoCols = staticmethod(mkTwoCols)
#-------------------------------------------------------------------------------
	def run(self):
		gtk.main()		
#===============================================================================
class Gui(Tkinter.Frame):
	def __init__(self, master):
		Tkinter.Frame.__init__(self, master)
		self.master.title("Address Book")
		self.ab = None
		self.pack()

		self.butQuit = Tkinter.Button(self)
		self.butQuit["text"] = "Quit"
		self.butQuit["command"] =  self.quit
		self.butQuit.pack({"side" : "bottom"})

		self.lbNames = Tkinter.Listbox(self, width=48, height=32)
		self.lbNames.pack({"side" : "top"})

	def doProperties(self):
		pass

	def setAddressBook(self, ab):
		self.ab = ab
		for names in self.ab.getAllNames().values():
			dispName = ""
			for name in names:
				if len(dispName) > 0:
					dispName += " "
				dispName += name
			self.lbNames.insert(Tkinter.END, dispName)
#===============================================================================		
class App:
#-------------------------------------------------------------------------------
	verbose = False
#-------------------------------------------------------------------------------
	def __init__(self):
		# Initialize Parameters
		self.doHelp = False
		self.showAllNames = False
		self.searchName = None
		self.importFileName = None
		self.ab = AddrBook()
		self.parseCmdLn()
		# If requested, do help. Otherwise, run app.
		if self.doHelp == True:
			self.showHelp()
		else:
			self.ab.connect()
			if self.searchName != None:
				self.doSearchName(self.searchName);
			elif self.showAllNames == True:
				self.doShowAllNames()
			elif self.importFileName != None:
				self.ab.importVcards(self.importFileName)
			else:
				if True:
					self.gui = GuiGtk(self.ab)
				else:
					self.doGui()
#-------------------------------------------------------------------------------
	def showHelp(self):
		helpText = "Options:\n" \
		         + "-help....................This text\n" \
		         + "-import vcardfilename....Import a vcard file\n" \
		         + "-dbFileName dbFileName...Use a different database\n" \
		         + "-searchName name.........Find an entity with specified name\n" \
		         + "-showAllNames............Display all names in database\n"
		print helpText
#-------------------------------------------------------------------------------
	def doGui(self):
		root = Tkinter.Tk()
		gui = Gui(root)
		gui.setAddressBook(self.ab)
		gui.mainloop()
		root.destroy()
#-------------------------------------------------------------------------------
	def parseCmdLn(self):
		argCnt = len(sys.argv)
		argIndex = 1
		while argIndex < argCnt:
			curOpt = sys.argv[argIndex]
			if cmp(curOpt, "-help") == 0:
				self.doHelp = True
			elif cmp(curOpt, "-verbose") == 0:
				App.verbose = True
			elif cmp(curOpt, "-searchName") == 0:
				argIndex += 1
				if argIndex == argCnt:
					raise Exception, "missing value for searchName option"
				self.searchName = sys.argv[argIndex]
			elif cmp(curOpt, "-dbFileName") == 0:
				argIndex += 1
				if argIndex == argCnt:
					raise Exception, "missing value for dbFileName option"
				self.ab.dbFileName = sys.argv[argIndex]
			elif cmp(curOpt, "-import") == 0:
				argIndex += 1
				if argIndex == argCnt:
					raise Exception, "missing filename for import option"
				self.importFileName = sys.argv[argIndex]
			elif cmp(curOpt, "-showAllNames") == 0:
				self.showAllNames = True
			else:
					raise Exception, "unknown option: " + curOpt
			argIndex += 1 # Next Option
#-------------------------------------------------------------------------------
	def doSearchName(self, name):
		for ent in self.ab.searchByName(name):
			print ent.getName()

			print "  Email Address(es)"
			for email in ent.emails:
				print "    " + email.toString()

			print "  Phone(s)"
			for phones in ent.phones:
				print "    " + phones.toString()
#-------------------------------------------------------------------------------
	def doShowAllNames(self):
		for names in self.ab.getAllNames().values():
			twoCols = GuiGtk.mkTwoCols(names)
			print twoCols[0],
			if len(twoCols[1]) > 0:
				print ",", twoCols[1]
			else:
				print
#===============================================================================		
if __name__ == "__main__":
	try:
		app = App()
	except Exception, e:
		print "Exception: " + str(e)
		if App.verbose:
			traceback.print_exc()

		

